


#ifndef CONNECTOR_H
#define CONNECTOR_H

#include "bip-time.h"
#include "structure.h"
#include "constraint.h"
#include "executable.h"
#include "priority.h"

#include <stdarg.h>
#include <ostream>
#include <map>

#ifdef SYMBOLIC
  #include "cudd.h"
#endif

class InteractionKey;
class Interaction;
class Compound;
class Port;

enum SynchronizationMode { SYNC_AND, SYNC_MAX, SYNC_MIN, SYNC_OR };

/*******************************
 * Class: Connector
 * Description: Connector class
 *******************************/
class Connector {
  /*** Private data members ***/
  private:
    static char LABEL[1024];

  /*** Protected data members ***/
  protected:
    char*                   mName;           /* name of connector */
    Compound*               mHolder;         /* holder compound */
    Array<Port, false>      mPorts;          /* port array */
    Port*                   mxPort;          /* exported port */
    bool                    mLegal;          /* legal i.e. complete or 
                                          maximal */
    bool                    mDominated;
    Array<Priority, false>  mDominatedPriorities; /* list of priorities in which the connector is dominated */
    Array<Priority, false>  mDominatorPriorities; /* list of priorities in which the connector is dominator */

    Interaction*            mFeasibleInter;  /* feasible interaction */
    bool                    mSubconn;        /* used as a sub-connector
                                          in hierarcical connector */
    bool                    mSubconnSameHolder; /* used as a sub-connector
					          in hierarcical connector
                                                  which is has the same holder */
    int                     mPriority;       /* static priority level */
    bool                    mQueued;         /* enqueued in the priority 
                                          heap */
    Connector*              mNext;           /* next legal */
    bool                    mEval;           /* flag marking evaluated */
    bool                    mRecomputexPort; /* recomputation of xPort() needed */
    Array<Connector, false> mLinks;     /* back pointer to higher links */

    mutable HashTable<InteractionKey,Interaction> mTable; /* table for storing interactions */


#ifdef SYMBOLIC
    DdNode*            mBdd;            /* BDD for Connector */
#endif

  /*** Constructors & Dectructors ***/
  public:
    Connector(const char*, 
              const Compound*, 
              const bool, 
              const int, ...);
    Connector(const char*,
              const Compound*,
              const bool);
    virtual ~Connector();

  /*** Public Utility Methods ***/
  protected:
           void add(Priority*);

  /*** Public Data-Access Methods ***/
  public:
    inline        char*               name()      const { return mName; }
    inline        Compound*           holder()    const { return mHolder; }

    inline const  Array<Port,     false>& ports()               const { return mPorts; }
    inline const  Array<Priority, false>& dominatedPriorities() const { return mDominatedPriorities; }
    inline const  Array<Priority, false>& dominatorPriorities() const { return mDominatorPriorities; }

    friend void Priority::setup();

    inline Port*               xPort()     const { return mxPort; }
    inline void                setxPort(Port* p) { mxPort = p; }
    inline bool                recomputexPort() const { return mRecomputexPort; }
    inline void                setRecomputexPort(const bool b) { mRecomputexPort = b; }
    inline bool                legal()     const { return mLegal; }
    inline bool                dominated() const { return mDominated; }
    inline void                setDominated(bool b) { mDominated = b; }

    inline Interaction*        feasibleInter() const
                                 { return mFeasibleInter; }                        
    inline void                setFeasibleInter(Interaction* i)
                                 { mFeasibleInter = i; }
    inline bool                isSubconn() const   { return mSubconn; }
    inline bool                isSubconnSameHolder() const { return mSubconnSameHolder; }
    inline void                setSubconnSameHolder(bool b) { mSubconnSameHolder = b; }
    inline int                 getPriority() const { return mPriority; }
    inline void                setPriority(int p)  { mPriority = p; }
    inline bool                noPrio()    const   
                                     { return (mPriority == -9999); }
    inline bool                isQueued()  const   { return mQueued; }
    inline void                setQueued(const bool b) { mQueued = b; }
    inline Connector*          next()      const   { return mNext; }
    inline void                setNext(Connector* next) 
                                 { mNext = next; }
    inline bool                evaluated() const   { return mEval; }
    inline void                setEvaluated(bool b) { mEval = b; }
    inline Array<Connector, false>& links() { return mLinks; }
    inline HashTable<InteractionKey, Interaction>&  table() const { return mTable; }

#ifdef SYMBOLIC
    inline DdNode*             bdd() const         { return mBdd; }
#endif
    virtual bool               isBroadcast() const { return true; }
            

   /*** Public Utility Methods ***/
   public:
     char*        label();
     char*        rt_label(const BitVector&);
     void         eval(bool opt);
     void         computexPort();
     Connector*   evaluate(Connector* last, bool isSubConn, bool opt);
     void         execute();
     void         execute(const BIP::Time&);
     void         actuate(const BIP::Time&);
     //void         notifyDependentPorts();
     void         createFeasibleInter();  /* create feasible interactions at the current state */
     void         createFeasibleInters(); /* create feasible interactions depending of the real-time (added by jcombaz) */
     Interaction* createFeasibleBroadcast(BitVector&, InteractionKey&, int, Interaction*);
     void         resetLink();     // reset this and dependents
     void         resetDependentLinks(); // reset dependents
     virtual bool                   complete() const;
     virtual bool                   guard();
     virtual bool                   guard(const BitVector&);
     virtual const Intervals&       constraint(const BitVector&);
             bool                   maximal() const;
     virtual void                   up();
     virtual void                   dn();
     virtual SynchronizationMode    syncMode(const BitVector&);

             BitVector              signature(int numPort, ...) const;

     void         setVarPorts(const int, va_list);
#ifdef SYMBOLIC
     void         initBdd();
     void         defineBdd(int, ...);
#endif

   /*** Public Methods invoked from appln. ***/
   public:
     virtual void define() {}
};


/*********************************
 * Class: Interaction
 * Description: Interaction class
 *********************************/
class InteractionKey {
  /*** Protected data members ***/
  protected:
    Connector* mConnector;
    BitVector  mSynced;

  /*** Constructors & Dectructors ***/
  public:
    InteractionKey(const InteractionKey& key):                     mConnector(key.connector()), mSynced(key.synced())     {}
    InteractionKey(Connector* connector):                          mConnector(connector),       mSynced(BitVector::EMPTY) {}
    InteractionKey(Connector* connector, const BitVector& synced): mConnector(connector),       mSynced(synced)           {}

  /*** Public Data-Access Methods ***/
  public:
    inline       Connector* connector() const { return mConnector; }
    inline const BitVector& synced()    const { return mSynced; }
    inline       BitVector& synced()          { return mSynced; }

    inline unsigned int     hash(unsigned int base)               const { return synced().hash(base); }
    inline bool             operator==(const InteractionKey& key) const { return synced() == key.synced(); }
};


/*********************************
 * Class: Interaction
 * Description: Interaction class
 *********************************/
class Interaction: public InteractionKey, public Executable {

  /*** Protected data members ***/
  protected:
         //SynchronizationMode   mSyncMode;   /* mode for synchronizing time constraints */
           Array<Priority,false> mInhibitors; /* potential inhibitors */

           Interaction*          mNext;        /* next legal interaction */


  /*** Constructors & Dectructors ***/
  public:
    Interaction(const InteractionKey&);
    Interaction(Connector*, const BitVector&);
    Interaction(const Interaction&);
    virtual ~Interaction();

  /*** Public Data-Access Methods ***/
  public:
    inline Array<Priority,false>& inhibitors()                      { return mInhibitors; }
    inline Interaction* next()                                const { return mNext; }
    inline void         setNext(Interaction* interaction)           { mNext = interaction; }

  /*** Public Utility Methods ***/
  public:
    const InteractionKey& key() const { return *this; }
    bool         equal(const Interaction*)    const;
    bool         isSubset(const Interaction*, 
                          bool&)              const;
    void         execute(const BIP::Time&);
    void         actuate(const BIP::Time&);
    //void         notifyDependentPorts();
    Interaction* copy()                       const;
    unsigned     hash(const unsigned)         const;

    void         print()                      const;
   
    static Interaction* Create(Connector*,  
                               const int, ...);
    static Interaction* Create(const InteractionKey&);
    static Interaction* CreateRealTime(const InteractionKey&);

    char*        label();

    friend std::ostream& operator<<(std::ostream&, const Interaction&);
};

#endif
