/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * engine.h
 *------------------------------------------------------------------------------
 *  BIP2 Real-Time Engine implementation
 *  + class Executables : list of Executable objects (i.e. internal port or
 *                                                     interactions)
 *  + class Engine      : Real-Time BIP2 Engine used for execution/exploration
 *                        of BIP2 models with time constraints
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Thu Dec 03 2009
 *
 ******************************************************************************/

#ifndef ENGINE_H
#define ENGINE_H

#include "constraint.h"
#include "executables.h"

#include <iostream>

#define SEMANTICS_LEVEL 0
#define SCHEDULER_LEVEL 1

class Compound;
class LogicalClock;
class PlatformClock;
class Compound;
class Executable;

/****************************************
 * Class: Filter
 * Description:
 *
 *
 ****************************************/
class Filter {
  /*** Protected data members ***/
  protected:
    unsigned int mLevel;

  /*** Constructors & Dectructors ***/
  public:
    Filter(unsigned int);
    virtual ~Filter();

  /*** Protected Data-Access Methods ***/
  protected:
    inline unsigned int   level()                         const { return mLevel; }
    inline void           setLevel(const unsigned int level)    { mLevel = level; }
    inline unsigned int   mask()                          const { return (1 << (level() + 1)) - 1; }
    inline unsigned int   mark()                          const { return (1 << level()); }

  /*** Public Utility Methods ***/
  public:
    virtual void filter(Executables&, Interval&) const = 0;
};


/**************************************************
 * Class: EmptySemantics
 * Description: no semantics applied
 **************************************************/
class EmptySemantics: public Filter {
  /*** Constructors & Dectructors ***/
  public:
    EmptySemantics();
    virtual ~EmptySemantics();

  /*** Public Utility Methods ***/
  public:
    virtual void filter(Executables&, Interval&) const;
};

/**************************************************
 * Class: StandardSemantics
 * Description: BIP standard semantics
 *  - waiting times are constrained by urgency
 *  - eliminates interactions that are not enabled
 *    between current time and next urgency
 **************************************************/
class StandardSemantics: public Filter {
  /*** Constructors & Dectructors ***/
  public:
    StandardSemantics();
    virtual ~StandardSemantics();

  /*** Public Utility Methods ***/
  public:
    virtual void filter(Executables&, Interval&) const;
};


/***********************************************
 * Class: EmptyScheduler
 * Description: no scheduling policy
 ***********************************************/
class EmptyScheduler: public Filter {
  /*** Constructors & Dectructors ***/
  public:
    EmptyScheduler();
    virtual ~EmptyScheduler();

  /*** Public Utility Methods ***/
  public:
    void filter(Executables&, Interval&) const;
};


/**************************************************
 * Class: EDFScheduler
 * Description: EDF Scheduler
 *  - minimal deadline interactions/internal ports
 *    have more priority
 *  - mask other interactions/internal ports
 **************************************************/
class EDFScheduler: public Filter {
  /*** Constructors & Dectructors ***/
  public:
    EDFScheduler();
    virtual ~EDFScheduler();

  /*** Public Utility Methods ***/
  public:
    virtual void filter(Executables&, Interval&) const;
};


/****************************************************
 * Class: EDFScheduler
 * Description: EDF-delta Scheduler
 *  - like EDFScheduler but non-urgent interactions/
 *    internal ports can be scheduled if there are
 *    delta time units between its next activation
 *    and the next urgency in the system
 ****************************************************/
class EDFdeltaScheduler: public Filter {
  /*** Protected data members ***/
  protected:
    BIP::Time mDelta;

  /*** Constructors & Dectructors ***/
  public:
    EDFdeltaScheduler(const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND));
    virtual ~EDFdeltaScheduler();

  /*** Public Data-Access Methods ***/
  protected:
    inline BIP::Time delta()               const { return mDelta; }
    inline void setDelta(const BIP::Time& delta) { mDelta = delta; }

  /*** Public Utility Methods ***/
  public:
    virtual void filter(Executables&, Interval&) const;
};


/**********************************************************
 * Class: Engine
 * Description:
 *   - generic engine used to animate BIP models
 *   - parameterized by the semantics (i.e. filter),
 *     logical and platform clocks
 *   - defines interfaces void initialize() and void run()
 **********************************************************/
class Engine {

  /*** Protected data members ***/
  protected:
    /* engine main parameters */
    Compound&      mSystem;        /* BIP system (main component) */
    LogicalClock&  mLogicalClock;  /* logical clock used for measuring model time */
    PlatformClock& mPlatformClock; /* platform clock used for measuring real-time */
    Filter&        mSemantics;     /* considered timed automata semantics */

    /* engine state */
    Executables    mInteractions;  /* list  of enabled interactions/internal ports */
    Interval       mWait;          /* range of enabled current time values */

    /* engine options and parameters */
    const bool           mQuiet;         /* no output if true */
    const unsigned int   mMaxSimLimit;   /* maximal number of executed interactions/internal ports (0 for unlimited) */
    std::ostream&        mSimout;        /* output stream */


  /*** Constructors & Dectructors ***/
  public:
    Engine(Compound&, LogicalClock&, Filter&, bool = true, unsigned int = 0, std::ostream& = std::cout);
    virtual ~Engine();


  /*** Protected Data-Access Methods ***/
  protected:
    inline       Compound&      system()                               { return mSystem; }
    inline       LogicalClock&  logicalClock()                         { return mLogicalClock; }
    inline       PlatformClock& platformClock()                        { return mPlatformClock; }
    inline       Filter&        semantics()                            { return mSemantics; }

    inline       bool           quiet()                          const { return mQuiet; }
    inline       unsigned int   maxSimLimit()                    const { return mMaxSimLimit; }
    inline       std::ostream&  simout()                         const { return mSimout; }

    inline       Executables&   interactions()                         { return mInteractions; }
    inline const Executables&   interactions()                   const { return mInteractions; }
    inline       Interval&      wait()                                 { return mWait; }

  /*** Public Data-Access Methods ***/
  public:


  /*** Protected Utility Methods ***/
  protected:
    bool deadlock() const { return interactions().empty(); }

  /*** Public Utility Methods ***/
  public:
    virtual void initialize()           = 0;
    virtual void run()                  = 0;
};


/**********************************************
 * Class: Executer
 * Description:
 *   - generic engine for executing BIP models
 *   - executers must define void choose()
 **********************************************/
class Executer: public Engine {

  /*** Protected data members ***/
  protected:
    /* executer main parameters */
    Filter&        mScheduler;

    /* executer choice using choose() */
    Executable*    mChoice;
    BIP::Time      mTime;

    /* executer options */
    BIP::Time      mMaxDrift;
    const int      mRandSeed;


  /*** Constructors & Dectructors ***/
  public:
    Executer(Compound&, LogicalClock&, Filter&, Filter&, bool = true, unsigned int = 0, std::ostream& = std::cout, int = 1, const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND));
    virtual ~Executer();


  /*** Protected Data-Access Methods ***/
  protected:
    inline Filter&        scheduler()              const { return mScheduler; }
    inline BIP::Time      maxDrift()               const { return mMaxDrift; }
    inline void setMaxDrift(const BIP::Time& maxDrift)   { mMaxDrift = maxDrift; }
    inline int            randSeed()               const { return mRandSeed; }

    inline Executable*    choice()                 const { return mChoice; }
    inline void           setChoice(Executable* choice)  { mChoice =  choice; }
    inline BIP::Time      time()                   const { return mTime; }
    inline void           setTime(const BIP::Time& time) { mTime = time; }


  /*** Public Data-Access Methods ***/
  public:
    virtual void         initialize();
    virtual void         run();
    virtual void         choose() = 0;  /* choose the next interaction/internal port to be executed */
};

/**********************************************
 * Class: OptimizedExecuter
 * Description:
 *   - optimized execution of interactions
 *   - everything is done in choose()
 **********************************************/

class OptimizedExecuter: public Executer {
  /*** Protected data members ***/
  protected:
    EmptySemantics mEmptySemantics;
    EmptyScheduler mEmptyScheduler;

    /* engine option and parameters */
    bool           mRandomized;

  /*** Constructors & Dectructors ***/
  public:
    OptimizedExecuter(Compound&, LogicalClock&, bool = true, unsigned int = 0, std::ostream& = std::cout, int = 1, const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND), bool = true);
    virtual ~OptimizedExecuter();

  /*** Protected Data-Access Methods ***/
  protected:
    inline bool randomized() const { return mRandomized; }

  /*** Public Utility Methods ***/
  public:
    void run();
    void choose();
};


/***************************************************
 * Class: StandardExecuter
 * Description:
 *   - choose the first interactions/internal ports
 *     amongst enabled ones and execute it ASAP
 ***************************************************/
class StandardExecuter: public Executer {

  /*** Constructors & Dectructors ***/
  public:
    StandardExecuter(Compound&, LogicalClock&, Filter&, Filter&, bool = true, unsigned int = 0, std::ostream& = std::cout, int = 1, const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND));
    virtual ~StandardExecuter();

  /*** Public Utility Methods ***/
  public:
    void choose();
};


/***********************************************************
 * Class: InteractiveExecuter
 * Description:
 *   - choice of waiting time and interaction/internal port
 *     is made by the user
 ***********************************************************/
class InteractiveExecuter: public Executer {
  /*** Constructors & Dectructors ***/
  public:
  InteractiveExecuter(Compound&, LogicalClock&, Filter&, Filter&, bool = true, unsigned int = 0, std::ostream& = std::cout, int = 1, const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND));
    virtual ~InteractiveExecuter();

  /*** Public Utility Methods ***/
  public:
    void choose();
};


/***********************************************************
 * Class: RandomExecuter
 * Description:
 *   - choice of waiting time and interaction/internal port
 *     is done randomly
 ***********************************************************/
class RandomExecuter: public Executer {

  /*** Constructors & Dectructors ***/
  public:
    RandomExecuter(Compound&, LogicalClock&, Filter&, Filter&, bool = true, unsigned int = 0, std::ostream& = std::cout, int = 1, const BIP::Time& = BIP::Time(40, BIP::Time::MILLISECOND));
    virtual ~RandomExecuter();


  /*** Protected Data-Access Methods */
  protected:
    int  randSeed()            const { return mRandSeed; }


  /*** Public Utility Methods ***/
  public:
    void choose();
};


/****************************************
 * Class: Explorer
 * Description:
 *
 *
 ****************************************/
class Explorer: public Engine {

  /*** Protected data members ***/
  protected:


  /*** Constructors & Dectructors ***/
  public:
    Explorer(Compound&, LogicalClock&, Filter&, bool);
    virtual ~Explorer();


  /*** Public Utility Methods ***/
  public:
    virtual void         initialize();
    virtual void         run();

};

#endif

