
#ifndef SCHEDULER_H
#define SCHEDULER_H

#include <fstream>

#include "structure.h"


extern bool BIP_ErrorStateReach ;
extern bool BIP_PruneStateReach ;

void BIP_ReachErrorState();
void BIP_ReachPruneState();


class Component;
class Connector;
class Node;
class Atomleaf;
class Leaf;

class Scheduler {

 protected:
  Compound*          mSystem;
  Connector*         mLegal;
  Port*              mInternals;
 
  PlatformClock*     mPlatformClock; 
  LogicalClock*      mLogicalClock;

  Connector*         mConnectorChoice;
  Interaction*       mChoice;
  Port*              mChoicePort;
  BIP::Time          mChoiceNext;
  BIP::Time          mChoiceDeadline;

  BIP::Time          mDeadline;

  BipHeap<Connector> mPrioHeap; /* priority-heap for 
                                   optimized execution */
  /* execution parameters */
  int  mMode;             // 0 => Normal, 1 => Interactive
  int  mIntQuiet;
  bool mQuiet;            // 1 => dont show scheduled interactions
  int  mRunMode;          // 0 => exploration, 1 => execution (trace), 2 => real time
  int  mTraverse;         // 0 => dfs exploration, 1=> bfs exploration
  int  mOptimizationMode; // 0 => Normal(old), 1 => max-synchronization
                          // 2 => fully prioritized

  bool mSimLimit;          // true => simulation run has a limit
  int  mMaxSimLimit;       // max no of steps(interactions) executed by engine
  int  mSimStep;           // count of steps executed by engine
  int  mDisable_output ;   // do not generate aut and atate files
  bool disable_output ;    // do not generate aut and atate files
  int  mTrace ;            // add trace in execution mode
  int  mRandSeed;          // seed for randomized execution


  FILE*   aut_file;        // Dumps the state-transitions 
  FILE*   state_file;      // Shows the state encodings
  std::ofstream simout_file; // Dumps the simulation trace output
  FILE*   pruned_paths;
  FILE*   deadlock_states;

  /* real-time parameters */
  int           mZeroWaiting;       // if true enforce zero-waiting execution of interactions
  BIP::Time     mMaxWait;           // maximal waiting time for random waits (non-zero-waiting execution of interactions)
  BIP::Time     mDrift;             // current drift between logical time and real-time
  BIP::Time     mMaxDrift;          // maximal allowed drift between logical time and real-time
  BIP::Time     mDelta;             // minimal time slot for executing non-urgent interactions (TIME_MAX for EDF execution)
  BIP::Time     mTolerance;         // tolerance to deadline misses (0 for no deadline miss)

 public:
  inline int   mode()             const { return mMode; }
  inline int   quiet()            const { return mQuiet; }
  inline int   runMode()          const { return mRunMode; }
  inline int   optimizationMode() const { return mOptimizationMode; }

  inline bool  simLimit()         const { return mSimLimit; } 
  inline int   maxSimLimit()      const { return mMaxSimLimit; }
  inline int   randSeed()         const { return mRandSeed; }
  std::ostream* simout;

 protected:
  Connector*   choose()        const;
  Connector*   userChoose()    const;
  void         rt_choose(const BIP::Time);

  inline PlatformClock* platformClock()   const { return mPlatformClock; }
  inline LogicalClock*  logicalClock()    const { return mLogicalClock; }

  inline Interaction* choice()            const { return mChoice; }
  inline Port*        choicePort()        const { return mChoicePort; }
  inline BIP::Time    choiceDeadline()    const { return mChoiceDeadline; }
  inline BIP::Time    choiceNext()        const { return mChoiceNext; }

  inline BIP::Time    deadline()          const { return mDeadline; }

  inline BIP::Time    drift()             const { return mDrift; }
  inline void         setDrift(BIP::Time drift) { mDrift = drift; }

  inline bool         zeroWaiting()       const { return mZeroWaiting; }
  inline BIP::Time    maxWait()           const { return mMaxWait; }

  void initDebugFct() const;
  void displayElement(char * cmd) const;
  void displayList(char * cmd) const;
  void setCurrent(char * cmd) const;

  void run_bfs_exploration(); // For BFS exploration
  void run_dfs_exploration(); // For DFS exploration
  void run_bfs_bis_exploration();
  void run_dfs_bis_exploration();

  void run_execution();   // For executing trace, with 
                          // following types:
    void runOptimizedExecution();  // Optimization 
                                   // with max-sync
    void runOptimizedExec2();      // Optimization with 
                                   // fully-prioritized connectors
    void runSinglestepExecution(); // Execution interactively, 
                                   // step-by-step

  bool execMaxSynchronize(Connector*); // execute with 
                                       // max-synchronization (optimization)
#ifdef SYMBOLIC
  void run_boolExecution();  // Execution with boolean connectors
#endif
  int ObserverState(Node*);
  /* Functions get interaction's path when ERROR */
  char* getLastChar(char*, char*);
  char* getFirstChar(char*, char*);
  char* fromState(char*);
  char* toState(char*);
  bool matchToState(char*, char*);
  void getInteraction(char* interaction, FILE* f, char* word, int line);
  void getPath(FILE* f);
    
 public:
  void run();
  void parseCmdFile(int argc, char** argv);  
 public:
  Scheduler(Compound*, 
            int = 0, 
            bool = false, 
            int = 1, 
            int = 0,
            int = 0);
  virtual ~Scheduler();

  inline BipHeap<Connector>& prioHeap() { return mPrioHeap; }

};


/*
 *
 */

class StateNode {
  protected:
    Node* stateInfo;
    int qtty;  // An integer quantity
    StateNode* next;

  public:
    StateNode();
    StateNode(Node*);
    ~StateNode();

    void setState(Node*);
    Node* getState();
    int getQtty();
    void setQtty(int);
    char* printState();
    StateNode* getNext();
    void setNext(StateNode*);
    bool equals(StateNode*);
};

/*
 *
 */

class StateNodeIter {
  protected:
    StateNode* head;
    StateNode* current;

  public:
    StateNodeIter();
    StateNodeIter(StateNode*);
    ~StateNodeIter();
    StateNode* getCurrent();
    StateNode* popHead();
    void addTail(StateNode*);
    void addHead(StateNode*);
    void addCurrent(StateNode*);
    StateNode* isPresent(StateNode*, int&);
    bool end();
    bool empty();
    void next();
};

#endif
